@@ -10,9 +10,7 @@
    as published by the Free Software Foundation; either version
    2 of the License, or (at your option) any later version.
 */
-/* $Id: ciped.c,v 1.16.4.2 1998/08/22 23:21:06 olaf Exp $ */
-
-#define CTLDIR "/etc/cipe"
+/* $Id: ciped.c,v 1.16.6.1 1998/11/14 20:17:21 olaf Exp $ */
 
 #include <errno.h>
 #include <fcntl.h>
@@ -22,6 +20,8 @@
 #include <string.h>
 #include <syslog.h>
 #include <unistd.h>
+#include <ctype.h>
+#include <mntent.h>
 #include <arpa/inet.h>
 #include <net/if.h>
 #include <sys/ioctl.h>
@@ -30,6 +30,14 @@
 #include <sys/stat.h>
 #include <sys/time.h>
 #include <sys/wait.h>
+#ifdef __GLIBC__      /* Ugh! the ehthernet address function needs this */
+#include <net/if_arp.h>
+#include <netinet/if_ether.h>
+#else
+#include <linux/if_ether.h>
+#include <linux/if_arp.h>
+#endif
+#include "linux/ipx.h"
 #include "ciped.h"
 #include "ioctl.h"
 
@@ -53,6 +61,7 @@ struct options {
 	int ovint;
 	char *ovstr;
 	struct sockaddr_in *ovsaddr;
+	struct sockaddr *ovhwaddr;
     } ov;
 } opts[] = {
 #define Opt(x,t) { #x, T##t, {0} },
@@ -62,6 +71,8 @@ struct options {
 #define Taddr	3
 #define Tuaddr	4
 #define Ttaddr	5
+#define Thwaddr 6
+#define Tipxnet 7
 
 #define Odevice		0  /* name of the CIPE device, e.g. cip3b0 */
 Opt(device,str)
@@ -101,6 +112,18 @@ Opt(arg,str)
 Opt(cttl,int)
 #define Omaxerr		17  /* Maximum ECONNREFUSED tolerance */
 Opt(maxerr,int)
+#define Onetmask        18  /* The device netmask address */
+Opt(netmask,addr)
+#define Ohwaddr         19  /* The device hardware address */
+Opt(hwaddr,hwaddr)
+#define Obroadcast      20  /* The device broadcast address */
+Opt(broadcast,addr)
+#define Obrdcstmode     21  /* An address for a static arp  for peer */
+Opt(brdcstmode,bool)       
+#define Oarphwaddr      22  /* the hardware address for the far end */
+Opt(arphwaddr,hwaddr)      
+#define Oipxnet         23  /* the CIPE links' IPX network address */
+Opt(ipxnet,ipxnet)
 
 #undef Opt
 { NULL, 0, {0} }};
@@ -110,6 +133,7 @@ Opt(maxerr,int)
 #define OA(x)      (opts[O##x].ov.ovsaddr)
 #define OAaddr(x)  (OA(x)->sin_addr)
 #define OAport(x)  (OA(x)->sin_port)
+#define OH(x)      (opts[O##x].ov.ovhwaddr)
 
 /* Statistics */
 struct kstat {
@@ -119,6 +143,65 @@ struct kstat {
 
 /*** Option parser ***/
 
+/* Parse a hardware address  - taken from ifconfig */
+int gethwaddr(char *c, struct sockaddr *sap)
+{
+    unsigned char *ptr;
+    char ch, *orig;
+    int i, val;
+
+    sap->sa_family = ARPHRD_ETHER;
+    ptr = sap->sa_data;
+
+    i = 0;
+    orig = c;
+    while((*c != '\0') && (i < ETH_ALEN)) {
+        val = 0;
+        ch = *c++;
+        if (isdigit(ch)) val = ch - '0';
+	else if (ch >= 'a' && ch <= 'f') val = ch - 'a' + 10;
+	else if (ch >= 'A' && ch <= 'F') val = ch - 'A' + 10;
+	else {
+	    fprintf(stderr, "gethwaddr: '%s' invalid hardware address\n",
+		    orig);
+	    return(-1);
+        }
+        val <<= 4;
+        ch = *c++;
+        if (isdigit(ch)) val |= ch - '0';
+	else if (ch >= 'a' && ch <= 'f') val |= ch - 'a' + 10;
+	else if (ch >= 'A' && ch <= 'F') val |= ch - 'A' + 10;
+	else {
+	    fprintf(stderr, "gethwaddr: '%s' invalid hardware address\n",
+		    orig);
+	    return(-1);
+        }
+        *ptr++ = (unsigned char) (val & 0377);
+        i++;
+
+        /* We might get a semicolon here - not required. */
+        if (*c == ':') {
+	    if (i == ETH_ALEN) {
+		; /* nothing */
+	    }
+	    c++;
+        }
+    }
+
+    return(0);
+}
+
+/* Convert an IPX address */
+int getipxnet(char *c, int *ipxnet) 
+{
+    if (sscanf(c, "%x", ipxnet)==0 || *ipxnet == 0) {
+	fprintf(stderr, "getipxnet: '%s' invalid IPX network address\n", c);
+	return -1;
+    }
+
+    return 0;
+}
+
 /* Resolve a host:port address */
 int getaddr(char *c, struct sockaddr_in *sa, const char *prot)
 {
@@ -131,7 +214,7 @@ int getaddr(char *c, struct sockaddr_in *sa, const char *prot)
     sa->sin_addr.s_addr=INADDR_ANY;
     if (q) {
 	if (!prot) {
-	    fprintf(stderr, "raw IP address has no port number\n");
+	    fprintf(stderr, "getaddr: raw IP address has no port number\n");
 	    return -1;
 	}
 	*q++='\0';
@@ -139,17 +222,17 @@ int getaddr(char *c, struct sockaddr_in *sa, const char *prot)
 	    if ((s=getservbyname(q, prot)))
 		sa->sin_port=s->s_port;
 	    else {
-		fprintf(stderr, "port '%s' invalid/not found\n", q);
+		fprintf(stderr, "getaddr: port '%s' invalid/not found\n", q);
 		return -1;
 	    }
 	}
     }
-    if ((sa->sin_addr.s_addr=inet_addr(c))==-1) {
+    if (!inet_aton(c, &sa->sin_addr)) {
 	if ((h=gethostbyname(c)))
 	    memcpy(&sa->sin_addr.s_addr, h->h_addr_list[0],
 		   sizeof(sa->sin_addr.s_addr));
 	else {
-	    fprintf(stderr, "host '%s' invalid/not found\n", c);
+	    fprintf(stderr, "getaddr: host '%s' invalid/not found\n", c);
 	    return -1;
 	}
     }
@@ -184,6 +267,23 @@ int setopt(const char *n, char *v)
 		    fprintf(stderr, "setopt: out of memory\n"); return -1;
 		}
 		if (getaddr(v, o->ov.ovsaddr, protos[o->otyp-Taddr])<0) {
+		    free(o->ov.ovsaddr);
+		    o->ov.ovsaddr = NULL;
+		    return -1;
+		}
+		break;
+	    case Thwaddr:
+		if (!(o->ov.ovhwaddr=malloc(sizeof(*o->ov.ovhwaddr)))) {
+		    fprintf(stderr, "setopt: out of memory\n"); return -1;
+		}
+		if (gethwaddr(v, o->ov.ovhwaddr)<0) {
+		    free(o->ov.ovhwaddr);
+		    o->ov.ovhwaddr = NULL;
+		    return -1;
+		}
+		break;
+	    case Tipxnet:
+		if(getipxnet(v, &o->ov.ovint) < 0) {
 		    return -1;
 		}
 		break;
@@ -265,6 +365,7 @@ void setopt_file(const char *n, int v)
 /* Print out all options */
 void dumpopt(void)
 {
+    char *hwaddr;
     struct options *o;
     for (o=opts; o->oname; ++o) {
 	printf("%s ", o->oname);
@@ -286,6 +387,27 @@ void dumpopt(void)
 	    else
 		printf("(none)\n");
 	    break;
+	case Thwaddr:
+	    if (o->ov.ovhwaddr) {
+		hwaddr = o->ov.ovhwaddr->sa_data;
+		printf("%02X:%02X:%02X:%02X:%02X:%02X\n",
+		       hwaddr[0],
+		       hwaddr[1],
+		       hwaddr[2],
+		       hwaddr[3],
+		       hwaddr[4],
+		       hwaddr[5]);
+	    }
+	    else
+		printf("(none)\n");
+	    break;
+	case Tipxnet:
+	    if (o->ov.ovint) {
+		printf("%08X\n", o->ov.ovint);
+	    }
+	    else
+		printf("(none)\n");
+	    break;
 	default: printf("internal error: %d\n", __LINE__); break;
 	}
     }
@@ -332,6 +454,195 @@ void argmiss(const char *s)
 #define dprintf1(f,a)  if(OI(debug)){printf(f,a);}
 
 
+/*** ARP code ***/
+
+/* set an ARP for the far end */
+int setarp (void) 
+{
+    struct arpreq arpreq;
+/*
+ * Sometime in the 1.3 series kernels, the arp request added a device name.
+ */
+#include <linux/version.h>
+#if LINUX_VERSION_CODE < 66381
+    char arpreq_arp_dev[32];
+#else
+#define  arpreq_arp_dev arpreq.arp_dev
+#endif
+    int sockfd = 0;
+
+#define SET_SA_FAMILY(addr, family)                     \
+    memset ((char *) &(addr), '\0', sizeof(addr));      \
+    addr.sa_family = (family);
+
+    memset (&arpreq, '\0', sizeof(arpreq));
+    
+    if (!OA(ptpaddr)) {
+	return 0;
+    }
+
+    if (!OH(arphwaddr)) {
+	return 0;
+    }
+    
+    if ((sockfd = socket(AF_INET,SOCK_DGRAM,0)) <0) {
+	perror("setarp: socket");
+	return -1;
+    }
+
+    memcpy(&arpreq.arp_ha, OH(arphwaddr), sizeof(*OH(arphwaddr)));
+    strncpy(arpreq_arp_dev, OS(device), sizeof(arpreq_arp_dev)-1);
+    SET_SA_FAMILY(arpreq.arp_pa, AF_INET);
+    ((struct sockaddr_in *) &arpreq.arp_pa)->sin_addr = OAaddr(ptpaddr);
+    arpreq.arp_flags = ATF_PERM;
+
+    if (ioctl(sockfd, SIOCSARP, (caddr_t)&arpreq) < 0) {
+	logerr(LOG_ERR, "setarp: SIOCSARP");
+	close(sockfd);
+	return -1;
+    }
+
+    close(sockfd);
+    return 0;
+}
+
+/*** IPX stuff ***/
+
+/* path_to_procfs - find the path to the proc file system mount point */
+
+char procpath[256];
+int ipxconfigured = 0;
+
+int path_to_procfs (void) {
+    struct mntent *mntent;
+    FILE *fp;
+  
+    fp = fopen (MOUNTED, "r");
+    if (fp != 0) {
+	mntent = getmntent (fp);
+	while (mntent != (struct mntent *) 0) {
+	    if (strcmp (mntent->mnt_type, MNTTYPE_IGNORE) != 0) {
+		if (strcmp (mntent->mnt_type, "proc") == 0) {
+		    strncpy (procpath, mntent->mnt_dir,
+			     sizeof (procpath)-10);
+		    procpath [sizeof (procpath)-10] = '\0';
+		    fclose (fp);
+		    return 1;
+		}
+	    }
+	    mntent = getmntent (fp);
+	}
+	fclose (fp);
+    }
+    return 0;
+}
+
+/* Set the IPX network number on the interface */
+int setipxnet(void) {
+    int    result = 1;
+
+    struct stat stat_buf;
+    int    skfd; 
+    struct ifreq         ifr;
+    struct sockaddr_ipx *sipx = (struct sockaddr_ipx *) &ifr.ifr_addr;
+
+    if (!OI(ipxnet)) {
+	/* Do nothing if no net number given */
+	return 0;
+    }
+
+    if (path_to_procfs()) {
+	strcat (procpath, "/net/ipx_interface");
+	if (lstat (procpath, &stat_buf) < 0) {
+	    logerr (LOG_ERR, "setipxnet: IPX support is not present in the kernel\n");
+	    return -1;
+	}
+    }
+
+    skfd = socket (AF_IPX, SOCK_DGRAM, 0);
+    if (skfd < 0) { 
+	if (errno != EIO) {
+	    logerr(LOG_DEBUG, "setipxnet: socket(AF_IPX)");
+	}
+	result = -1;
+    }
+    else {
+	memset (&ifr, '\0', sizeof (ifr));
+	strcpy (ifr.ifr_name, OS(device));
+
+	memcpy (sipx->sipx_node, OH(hwaddr)->sa_data, IPX_NODE_LEN);
+	sipx->sipx_family  = AF_IPX;
+	sipx->sipx_port    = 0;
+	sipx->sipx_network = htonl (OI(ipxnet));
+	sipx->sipx_type    = IPX_FRAME_ETHERII;
+	sipx->sipx_action  = IPX_CRTITF;
+	/*
+	 *  Set the IPX device
+	 */
+	if (ioctl(skfd, SIOCSIFADDR, (caddr_t) &ifr) < 0) {
+	    result = -1;
+	    if (errno != EEXIST) {
+		if (errno != EIO) {
+		    logerr (LOG_DEBUG,
+			    "setipxnet: ioctl(SIOCAIFADDR, CRTITF)");
+		}
+	    }
+	    else {
+		logerr (LOG_WARNING,
+			"setipxnet: ioctl(SIOCAIFADDR, CRTITF): Address already exists");
+	    }
+	}
+    
+	ipxconfigured = 1;
+
+	close (skfd);
+    }
+    return result;
+}
+
+/* Clear the IPX network information */
+int clearipxnet (void) {
+    int    result = 0;
+    int    skfd; 
+    struct ifreq         ifr;
+    struct sockaddr_ipx *sipx = (struct sockaddr_ipx *) &ifr.ifr_addr;
+
+    if( !ipxconfigured ) {
+	/* Do not attmept this if IPX is not configured */
+	return 0;
+    }
+
+    skfd = socket (AF_IPX, SOCK_DGRAM, 0);
+    if (skfd < 0){ 
+	if (errno != EIO) {
+	    logerr(LOG_DEBUG, "clearipxnet: socket(AF_IPX)");
+	}
+	result = -1;
+    }
+    else {
+	memset (&ifr, '\0', sizeof (ifr));
+	strcpy (ifr.ifr_name, OS(device));
+
+	sipx->sipx_type    = IPX_FRAME_ETHERII;
+	sipx->sipx_action  = IPX_DLTITF;
+	sipx->sipx_family  = AF_IPX;
+/*
+ *  Set the IPX device
+ */
+	if (ioctl(skfd, SIOCSIFADDR, (caddr_t) &ifr) < 0) {
+	    if (errno != EIO) {
+		logerr (LOG_INFO, "clearipxnet: ioctl(SIOCAIFADDR, IPX_DLTITF)");
+	    }
+	    result = -1;
+	}
+	close (skfd);
+    }
+
+    return result;
+}
+
+
+
 /*** Initialize, open and attach the CIPE device. ***/
 
 int opendev(void)
@@ -340,7 +651,7 @@ int opendev(void)
     struct siocsifcippar p;
     struct siocsifcipkey k;
     struct siocsifcipatt a;
-    struct ifreq r;
+    struct ifreq r,s,t;
 
 #define amiss(s) { argmiss(s); goto error; }
 #define err(s)   { logerr(LOG_ERR, s); goto error; }
@@ -420,17 +731,59 @@ int opendev(void)
 	err("opendev: attach");
 
     /* Perform ifconfig */
+    memset((char *)&r, 0, sizeof(r));
+    memset((char *)&s, 0, sizeof(s));
+    memset((char *)&t, 0, sizeof(t));
     strcpy(r.ifr_name, OS(device));
+    strcpy(s.ifr_name, OS(device));
+    if (OH(hwaddr)) {
+	memcpy(&r.ifr_hwaddr, OH(hwaddr), sizeof(*OH(hwaddr)));
+	if (ioctl(f, SIOCSIFHWADDR, &r)<0)
+	    err("opendev: SIOCSIFHWADDR");
+      
+    }
     if (OA(ipaddr)) {
 	memcpy(&r.ifr_addr, OA(ipaddr), sizeof(*OA(ipaddr)));
 	if (ioctl(f, SIOCSIFADDR, &r)<0)
 	    err("opendev: SIOCSIFADDR");
     } else
 	amiss("ipaddr");
-    if (OA(ptpaddr)) {
+    if (!OI(brdcstmode)) {
+	/* point to point mode - MUST have a point to point address */
+	if (!OA(ptpaddr)) {
+	    amiss("ptpaddr");
+	}
+	s.ifr_flags |= IFF_POINTOPOINT;
 	memcpy(&r.ifr_dstaddr, OA(ptpaddr), sizeof(*OA(ptpaddr)));
 	if (ioctl(f, SIOCSIFDSTADDR, &r)<0)
 	    err("opendev: SIOCSIFDSTADDR");
+	t = s;
+	if (ioctl(f, SIOCSIFFLAGS, &t)<0)
+	    err("opendev: SIOCSIFFLAGS");
+    } 
+    else {
+	/* Broadcast mode - set IFF flags */ 
+	s.ifr_flags |= IFF_BROADCAST;
+#ifdef CIPE_MULTICAST      
+	s.ifr_flags |= IFF_MULTICAST;
+#endif
+	t = s;
+	if (ioctl(f, SIOCSIFFLAGS, &t)<0)
+	    err("opendev: SIOCSIFFLAGS");
+    }
+    if (OA(broadcast)) {
+	s.ifr_flags |= IFF_BROADCAST;
+	memcpy(&r.ifr_broadaddr, OA(broadcast), sizeof(*OA(broadcast)));
+	if (ioctl(f, SIOCSIFBRDADDR, &r)<0)
+	    err("opendev: SIOCSIFBRDADDR");
+	t = s;
+	if (ioctl(f, SIOCSIFFLAGS, &t)<0)
+	    err("opendev: SIOCSIFFLAGS");
+    }
+    if (OA(netmask)) {
+	memcpy(&r.ifr_netmask, OA(netmask), sizeof(*OA(netmask)));
+	if (ioctl(f, SIOCSIFNETMASK, &r)<0)
+	    err("opendev: SIOCSIFNETMASK");
     }
     if (OI(mtu)) {
 	r.ifr_mtu=OI(mtu);
@@ -442,13 +795,19 @@ int opendev(void)
 	if (ioctl(f, SIOCSIFMETRIC, &r)<0)
 	    err("opendev: SIOCSIFMETRIC");
     }
-    r.ifr_flags=IFF_UP|IFF_RUNNING|IFF_NOARP|IFF_NOTRAILERS|IFF_POINTOPOINT;
-    if (ioctl(f, SIOCSIFFLAGS, &r)<0)
+    s.ifr_flags |= IFF_UP|IFF_RUNNING|IFF_NOTRAILERS;
+    if (ioctl(f, SIOCSIFFLAGS, &s)<0)
 	err("opendev: SIOCSIFFLAGS");
 
+    /* Not setting the ARP for the far end of the tunnel is not fatal */
+    setarp();
+
+    /* Set up IPX network */
+    setipxnet();
+
     return f;
 
- error:
+error:
     close(f);
     return -1;
 
@@ -691,9 +1050,6 @@ int kxchg(int f, int r)
     if (x<0) {
 	errno=e;
 	logerr(LOG_WARNING, "kxchg: send");
-	if (e==ENXIO)
-	    return -1;
-	x=(OI(maxerr)<0 ? 0 : --errcnt);
     }
     return x;
 }
@@ -711,13 +1067,22 @@ void pspawn(char *s)
     if (c) {
 	(void) wait(NULL); /* exactly one child is active */
     } else {
-	/* call <interface> <my-addr> <daemon-pid> <local> <remote> <arg> */
-	char bufa[32], bufb[32], bufc[32], bufd[32];
-	char *na[]={s, OS(device), bufa, bufb, bufc, bufd, OS(arg), NULL};
+      	/* call <interface> <my-addr> <daemon-pid> <local> <remote> <network-address> <netmask> <arg>*/
+        struct in_addr networkaddr;
+	struct in_addr netmask;
+	char bufa[32], bufb[32], bufc[32], bufd[32], bufe[32], buff[32];
+	char *na[]={s, OS(device), bufa, bufb, bufc, bufd, 
+		    bufe, buff, OS(arg), NULL};
+
+	netmask.s_addr = (OA(netmask) ?  OAaddr(netmask).s_addr : 0xffffffff);
+	networkaddr.s_addr = htonl(ntohl(OAaddr(ipaddr).s_addr) 
+				   & ntohl(OAaddr(netmask).s_addr));
 	sprintf(bufa, "%s:%d", inet_ntoa(OAaddr(me)), ntohs(OAport(me)));
 	sprintf(bufb, "%d", p);
 	sprintf(bufc, "%s", inet_ntoa(OAaddr(ipaddr)));
 	sprintf(bufd, "%s", inet_ntoa(OAaddr(ptpaddr)));
+	sprintf(bufe, "%s", inet_ntoa(networkaddr));
+	sprintf(buff, "%s", inet_ntoa(netmask));
 	close(0); close(1); close(2);
 	(void) open("/dev/null", O_RDWR);
 	(void) dup(0); (void) dup(0);
@@ -813,6 +1178,9 @@ int main(int argc, char *argv[])
 	}
     }
 
+    /* Clear ipx configuration */
+    clearipxnet();
+
     closedev(fd);
     if (OI(debug))
 	printf("KX stats: rreq=%d, req=%d, ind=%d, indb=%d, ack=%d,"
